# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

from cstar.exceptions import UnknownHost

Host = namedtuple("Host", "fqdn ip dc cluster token is_up")
Host.__hash__ = lambda self: self.ip.__hash__()


def _host_eq(self, other):
    if hasattr(other, 'ip'):
        return self.ip == other.ip
    return self.ip == other


Host.__eq__ = _host_eq


class Topology(object):
    """An immutable type describing a topology of C* clusters. Contains utility methods for creating filtered
    subtopologies.

    This type is meant to be used without mutating it."""

    def __init__(self, hosts=[]):
        self.hosts = set(hosts)

    def first(self):
        """Return first host in topology (by cluster position)"""
        if not self:
            return None
        return sorted(self.hosts, key=lambda x: x.token)[0]

    def get_host(self, ip):
        for h in self:
            if h.ip == ip:
                return h
        raise UnknownHost(ip)

    def with_cluster(self, cluster):
        """Return subtopology filtered on cluster"""
        return Topology(filter(lambda host: cluster == host.cluster, self.hosts))

    def with_dc(self, dc):
        """Return subtopology filtered on dc"""
        return Topology(filter(lambda host: dc == host.dc, self.hosts))

    def without_host(self, host):
        """Return subtopology without specified host"""
        return Topology(self.hosts - set((host,)))

    def without_hosts(self, hosts):
        """Return subtopology without specified hosts"""
        return Topology(self.hosts - set(hosts))

    def get_clusters(self):
        """Returns a set containing all the individual clusters in this topology"""
        return set(host.cluster for host in self.hosts)

    def get_dcs(self):
        """Returns a set containing all the individual data centers in this topology"""
        return set(host.dc for host in self.hosts)

    def get_down(self):
        """Returns a set of all nodes that are down in this topology"""
        return Topology(node for node in self if not node.is_up)

    def get_up(self):
        """Returns a set of all nodes that are up in this topology"""
        return Topology(node for node in self if node.is_up)

    def __contains__(self, o):
        return self.hosts.__contains__(o)

    def __str__(self):
        return " ".join(host.fqdn for host in self.hosts)

    def __or__(self, other):
        return Topology(self.hosts | other.hosts)

    def __len__(self):
        return len(self.hosts)

    def __iter__(self):
        return iter(self.hosts)

    def __repr__(self):
        return "Topology(%s)" % (self.hosts,)

    def __eq__(self, other):
        return self.hosts.__eq__(other.hosts)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import signal
import sys

import cstar.jobwriter
from cstar.output import emph

_msg = None
_job = None


def _handler(signum, frame):
    if cstar.jobwriter.write(_job):
        print("\n" + _msg)
    sys.exit(1)


def print_message_and_save_on_sigint(job, job_id):
    global _msg, _job
    _msg = "Shutting down gracefully. Hit ^C again to shut down gracelessly.\n\nTo resume, type %s" % (
        emph("cstar continue " + job_id),)

    _job = job
    signal.signal(signal.SIGINT, _handler)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Argument parsing"""


def _add_common_arguments(parser):
    parser.add_argument('--stop-after', type=int, help='Stop the job after specified number of hosts')
    parser.add_argument('--verbose', '-v', action="count", default=0, help='Increase command output verbosity')
    parser.add_argument('--output-directory', help='Output location for job log')
    parser.add_argument('--ignore-down-nodes', action="store_true", default=False,
                        help='Run the command even if there are down nodes in the cluster')


def _add_destination_arguments(parser):
    parser.add_argument('--seed-host', nargs='*',
                        help='One or more hosts to use as seeds for the cluster topology')

    parser.add_argument('--host', nargs='*', help='One or more hosts to run the script on')

    parser.add_argument('--host-file',
                        help='A file containing one or more hosts to run the script on (newline separated)')
    # The amount of time in seconds to sleep after spinning up a new runner.
    # Spinning up too many things at once seems to cause timeouts. Not sure
    #  if it's paramiko being slow or what, but this seems to help.
    parser.add_argument('--ssh-pause-time', type=float, default=0.5,
                        help='The amount of time to pause between establishing new ssh connections to avoid timeouts')
    parser.add_argument('--node-done-pause-time', type=float, default=0.0,
                        help='The amount of time (in seconds) to pause between a node has finished and the next node starting')


def _add_strategy_arguments(parser):
    parser.add_argument('--max-concurrency', '-j', type=int,
                        help='Maximum number of hosts to run the job on concurrently')

    parser.add_argument('--timeout', type=int, default=None,
                        help='Maximum number of seconds to run on one host before considering the job failed')

    parser.add_argument('--strategy', choices=("one", "topology", "all"), default=None,
                        help='Maximum number of seconds to run on one host before considering the job failed')

    parser.add_argument('--cluster-parallel', action="store_true", default=None, help='Run on all clusters in parallel')
    parser.add_argument('--cluster-serial', dest='cluster_parallel', action="store_false",
                        help='Run on all clusters in serial')
    parser.add_argument('--dc-parallel', action="store_true", default=None,
                        help='Run on all data centers of a cluster in parallel')
    parser.add_argument('--dc-serial', dest='dc_parallel', action="store_false",
                        help='Run on all data centers of a cluster in serial')
    parser.add_argument('--dc-filter', type=str, default=None,
                        help='Only run in hosts belonging to the specified data center')
    parser.add_argument('--key-space', '--keyspace',
                        help='The keyspace to use for endpoint mapping calculation. Uses all keypsaces by default.')


def _add_cstar_arguments_without_command(parser):
    """Argument parsing for case when cstar is called without specifying a command to run"""
    parser.add_argument('--max-job-age', default=7, type=int, help='Maximum age in days of a job to resume')


def add_cstar_arguments(parser, commands, execute_command, execute_continue, execute_cleanup):
    """Argument parsing for case when cstar is called specifying a command to run"""
    subparsers = parser.add_subparsers(dest='sub_command')

    continue_parser = subparsers.add_parser('continue', help='Continue a previously created job (*)')
    continue_parser.set_defaults(func=execute_continue)
    continue_parser.add_argument('job_id')
    continue_parser.add_argument('--retry-failed', action="store_true", default=False,
                        help='Retry failed nodes.')
    _add_common_arguments(continue_parser)
    _add_cstar_arguments_without_command(continue_parser)
    _add_ssh_arguments(continue_parser)

    cleanup_parser = subparsers.add_parser('cleanup-jobs', help='Cleanup old finished jobs and exit (*)')
    cleanup_parser.set_defaults(func=execute_cleanup)
    _add_common_arguments(cleanup_parser)
    _add_cstar_arguments_without_command(cleanup_parser)
    _add_ssh_arguments(cleanup_parser)

    for (name, command) in commands.items():
        command_parser = subparsers.add_parser(name, help=command.description)
        for arg in command.arguments:
            command_parser.add_argument(arg.option, dest=arg.name, help=arg.description, required=arg.required, default=arg.default)
        _add_destination_arguments(command_parser)
        _add_strategy_arguments(command_parser)
        _add_common_arguments(command_parser)
        _add_ssh_arguments(command_parser)
        command_parser.set_defaults(func=lambda args: execute_command(args), command=command)


def add_cstarpar_arguments(parser):
    """Argument parsing for cstarpar"""
    _add_destination_arguments(parser)
    _add_common_arguments(parser)
    _add_strategy_arguments(parser)
    _add_ssh_arguments(parser)
    parser.add_argument('command', help='Command to run once for each Cassandra host')

def _add_ssh_arguments(parser):
    parser.add_argument('--ssh-username', help='Username for ssh connection', default=None)
    parser.add_argument('--ssh-password', help='Password for ssh connection', default=None)
    parser.add_argument('--ssh-identity-file', help='Identity file for ssh connection', default=None)
    


# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import datetime

import cstar.strategy
import cstar.job
import cstar.state
import cstar.progress
import cstar.topology

FILE_FORMAT_VERSION = 6


def _to_dict(child):
    if type(child) is set:
        return list(child)
    if type(child) is cstar.topology.Topology:
        return list(child)
    if type(child) is cstar.strategy.Strategy:
        return cstar.strategy.serialize(child)
    if type(child) is cstar.state.State:
        return _state_to_dict(child)
    if type(child) is cstar.progress.Progress:
        return _progress_to_dict(child)
    if type(child) is cstar.job.Job:
        return _job_to_dict(child)
    return child


def _state_to_dict(self):
    skip = {"endpoint_mapping", "stop_after"}
    data = dict((key, _to_dict(val)) for key, val in self.__dict__.items() if
                not key.startswith('_') and key not in skip)
    return data


def _progress_to_dict(self):
    skip = {}
    data = dict((key, _to_dict(val)) for key, val in self.__dict__.items() if
                not key.startswith('_') and key not in skip)
    return data


def _job_to_dict(self):
    skip = {"results", "handled_finished_jobs", "do_loop", "job_id", "job_runner"}
    data = dict((key, _to_dict(val)) for key, val in self.__dict__.items() if key[0] != '_' and key not in skip)
    data["version"] = FILE_FORMAT_VERSION
    data["job_runner"] = self.job_runner.__name__
    data["creation_timestamp"] = int(datetime.datetime.utcnow().timestamp())
    return data


def _job_to_json(job):
    return json.dumps(_to_dict(job), sort_keys=True, indent=4)


def write(job):
    if not job.state:
        return False
    with open(job.output_directory + "/job.json", 'w') as f:
        f.write(_job_to_json(job) + "\n")
    return True



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parses the output of 'nodetool describering' and returns a simple AST

Standard parser implemented in two phases (tokenization and parsing).
The parser is a rather simple recursive descent affair.

BNF grammar:

root_expressions ::= root_expression '\n' root_expressions | root_expression

root_expression ::= identifiers ':' identifiers_or_block

identifiers_or_block ::= identifiers | BEGIN expressions END

identifiers ::= identifier identifiers | identifier

expressions ::= expression '\n' expressions | expression

expression ::= call | list | identifier

call ::= identifier '(' opt_arguments ')'

opt_arguments ::= | arguments

arguments ::= argument ',' arguments | argument

argument ::= identifier ':' expression

list ::= '[' opt_expressions ']'

opt_expressions ::= | expressions

expressions ::= expression ',' expressions | expression
"""

from collections import namedtuple
from cstar.exceptions import ParseException
import re

Call = namedtuple("Call", "name arguments")


class Symbol(object):
    def __init__(self, s, pos):
        self.val = s
        self.offset = pos


class Identifier(object):
    def __init__(self, s, pos):
        self.val = s
        self.offset = pos


def parse(text):
    lines = list(filter(lambda x: x, [_parse_line(line) for line in text.split('\n')]))
    return lines


def _parse_line(line):
    tokens = list(reversed(_tokenize(line)))
    if not tokens:
        return None

    if len(tokens) >= 2 and type(tokens[-1]) is Identifier and tokens[-2].val == '(':
        return _parse_call(line, tokens)

    return None


def _parse_call(line, tokens):
    name = tokens.pop()
    res = {}

    if not type(name) is Identifier:
        raise ParseException(line, name.offset, "Bad identifier")

    paran = tokens.pop()
    if paran.val != '(':
        raise ParseException(line, paran.offset, "Expected '('")

    while True:
        _parse_argument(line, tokens, res)
        if tokens[-1].val == ',':
            tokens.pop()
        else:
            break
    if tokens[-1].val != ')':
        raise ParseException(line, tokens[-1].offset, "Expected a ')'")
    tokens.pop()
    return Call(name.val, res)


def _parse_argument(line, tokens, res):
    if not tokens:
        raise ParseException(line, len(line), "Premature end of line")

    name = tokens.pop()
    if not type(name) is Identifier:
        raise ParseException(line, name.offset, "Expected an identifier, not a symbol")

    sym = tokens.pop()
    if sym.val != ':':
        raise ParseException(line, sym.offset, "Expected ':'")

    value = _parse_expression(line, tokens)
    res[name.val] = value


def _parse_expression(line, tokens):
    if tokens[-1].val == '[':
        return _parse_describe_line_list(line, tokens)

    if type(tokens[-1]) is Identifier:
        if (len(tokens) >= 2) and (tokens[-2].val == '('):
            return _parse_call(line, tokens)
        else:
            return tokens.pop().val
    raise ParseException(line, tokens[-1].offset, "Expected identifier, call or list, got '%s'")


def _parse_describe_line_list(line, tokens):
    tokens.pop()
    res = []
    if tokens[-1].val != ']':
        while True:
            res.append(_parse_expression(line, tokens))
            if tokens[-1].val == ',':
                tokens.pop()
            elif tokens[-1].val == ']':
                break
            else:
                raise ParseException(line, tokens[-1].offset, "Expected ',' or ']', got '%s' in line %s")

    tokens.pop()
    return res


def _tokenize(original_line, offset=0):
    line = original_line.lstrip()
    offset += len(original_line) - len(line)
    if not line:
        return ()
    if line[0] in {'(', ')', '[', ']', ',', ':'}:
        return (Symbol(line[0], offset),) + _tokenize(line[1:], offset + 1)
    if line[0].isalnum() or line[0] in {'-'}:
        tok = ""
        while True:
            tok += line[0]
            line = line[1:]
            if not line:
                break
            if not (line[0].isalnum() or line[0] in {'.', '_', '-'}):
                break

        if tok.isnumeric() or (tok[0] == '-' and tok[1:].isnumeric()):
            identifier = Identifier(int(tok), offset)
        else:
            identifier = Identifier(tok, offset)
        return (identifier,) + _tokenize(line, offset + len(tok))
    raise ParseException(line, offset, "Could not parse string")

def convert_describering_to_range_mapping(tokens):
    range_mapping = []
    for token in tokens:
        range = {}
        range["startToken"] = str(token.arguments['start_token'])
        range["endToken"] = str(token.arguments['end_token'])
        range["endpoints"] = token.arguments['endpoints']
        range_mapping.append(range)
    
    return range_mapping



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import re

from cstar.topology import Topology, Host

_cluster_name_re = re.compile(r"^\s*Name:\s*(.*)$", re.MULTILINE)
_ip_re = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
_token_re = re.compile(r"^\-?\d+$")
_status_re = re.compile(r"^[A-Za-z]+$")
_keyspace_name_re = re.compile(r"^\s*Keyspace:\s*(.*)$", re.MULTILINE)


def parse_describe_cluster(text):
    return _cluster_name_re.search(text).group(1)


def _parse_node(line):
    words = line.split()
    if len(words) == 8 and re.match(_ip_re, words[0]) and re.match(_status_re, words[2]) and re.match(_token_re, words[7]):
        return words
    else:
        return None


def _parse_datacenter_name_and_nodes(datacenter_section):
    lines = datacenter_section.split("\n")
    name = lines[0]
    nodes = [_parse_node(line) for line in lines[1:]]
    return (name, [node for node in nodes if node is not None])


def parse_nodetool_ring(text, cluster_name, reverse_dns_preheat):
    topology = []
    datacenter_sections = text.split("Datacenter: ")[1:]
    datacenter_names_and_nodes = [_parse_datacenter_name_and_nodes(section) for section in datacenter_sections]
    reverse_dns_preheat([node[0] for (_, nodes) in datacenter_names_and_nodes for node in nodes])
    for (datacenterName, nodes) in datacenter_names_and_nodes:
        for node in nodes:
            fqdn = node[0]
            try:
                fqdn=socket.gethostbyaddr(node[0])[0]
            except socket.herror:
                pass
            topology.append(Host(fqdn=fqdn, ip=node[0], dc=datacenterName, cluster=cluster_name,
                                 is_up=node[2] == "Up", token=int(node[7])))

    return Topology(topology)


def extract_keyspaces_from_cfstats(text):
    return re.findall(_keyspace_name_re, text)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cstar.output

"""Pretty-print the progress of a cstar job"""


def print_progress(original_topology, progress, down, printer=cstar.output.print_topology):
    def get_status(host):
        if host in progress.done:
            if host in down:
                return "-"
            return '+'
        if host in progress.running:
            if host in down:
                return "/"
            return '*'
        if host in progress.failed:
            if host in down:
                return "X"
            return '!'
        if host in down:
            return ":"
        return '.'

    lines = [" +  Done, up      * Executing, up      !  Failed, up      . Waiting, up",
             " -  Done, down    / Executing, down    X  Failed, down    : Waiting, down"]

    clusters = sorted(original_topology.get_clusters())
    for cluster in clusters:
        if len(clusters):
            lines.append("Cluster: " + cluster)
        cluster_topology = original_topology.with_cluster(cluster)
        dcs = sorted(cluster_topology.get_dcs())
        for dc in dcs:
            if len(dcs):
                lines.append("DC: " + dc)
            dc_topology = cluster_topology.with_dc(dc)
            hosts = sorted(dc_topology, key=lambda x: x.token)
            status = "".join([get_status(host) for host in hosts])
            if len(status) >= 6:
                status = status[0:len(status):3] + "\n" + status[1:len(status):3] + "\n" + status[2:len(status):3]
            lines.append(status)
    lines.append("%d done, %d failed, %d executing" % (len(progress.done), len(progress.failed), len(progress.running)))
    printer("\n".join(lines))



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Argument parsing and related plumbing for the cstar command"""

import argparse
import sys
import uuid

import cstar.args
import cstar.command
import cstar.strategy
import cstar.job
import cstar.remote
import cstar.jobreader
from cstar.output import msg, error, emph
import cstar.output
from cstar.exceptions import BadFileFormatVersion, FileTooOld,NoHostsSpecified
import cstar.signalhandler
import cstar.jobrunner
import cstar.cleanup


def fallback(*args):
    for arg in args:
        if arg is not None:
            return arg


def get_commands():
    result = {}
    names = cstar.command.list()
    for name in sorted(names):
        sub = cstar.command.load(name)
        result[name] = sub
    return result


def execute_continue(args):
    msg("Retry : ", args.retry_failed)
    with cstar.job.Job() as job:
        try:
            cstar.jobreader.read(job, args.job_id, args.stop_after, max_days=args.max_job_age,
                                 output_directory=args.output_directory, retry=args.retry_failed)
        except (FileTooOld, BadFileFormatVersion) as e:
            error(e)
        msg("Resuming job", job.job_id)
        msg("Running ", job.command)

        cstar.signalhandler.print_message_and_save_on_sigint(job, job.job_id)

        job.resume()


def execute_cleanup(args):
    msg('Cleaning up old jobs')
    cstar.cleanup.cleanup(args.max_job_age)


def execute_command(args):
    cstar.output.debug(args)
    command = args.command
    if bool(args.seed_host) + bool(args.host) + bool(args.host_file) != 1:
        error("Exactly one of --seed-host, --host and --host-file must be used", print_traceback=False)

    hosts = None

    if args.host_file:
        with open(args.host_file) as f:
            hosts = f.readlines()

    if args.host:
        hosts = args.host

    with cstar.job.Job() as job:
        env = dict((arg.name, getattr(args, arg.name)) for arg in command.arguments)
        job_id = str(uuid.uuid4())
        msg("Job id is", emph(job_id))
        msg("Running", command.file)

        cstar.signalhandler.print_message_and_save_on_sigint(job, job_id)

        job.setup(
            hosts=hosts,
            seeds=args.seed_host,
            command=command.file,
            job_id=job_id,
            strategy=cstar.strategy.parse(fallback(args.strategy, command.strategy, "topology")),
            cluster_parallel=fallback(args.cluster_parallel, command.cluster_parallel, False),
            dc_parallel=fallback(args.dc_parallel, command.dc_parallel, False),
            max_concurrency=args.max_concurrency,
            timeout=args.timeout,
            env=env,
            stop_after=args.stop_after,
            job_runner=cstar.jobrunner.RemoteJobRunner,
            key_space=args.key_space,
            output_directory=args.output_directory,
            ignore_down_nodes=args.ignore_down_nodes,
            dc_filter=args.dc_filter,
            sleep_on_new_runner=args.ssh_pause_time,
            sleep_after_done=args.node_done_pause_time,
            ssh_username = args.ssh_username,
            ssh_password = args.ssh_password,
            ssh_identity_file = args.ssh_identity_file)
        job.run()


def main():
    parser = argparse.ArgumentParser(
        description='cstar', prog='cstar', formatter_class=argparse.RawDescriptionHelpFormatter, epilog="(*): Special built-in cstar job management action")
    cstar.args.add_cstar_arguments(parser, get_commands(), execute_command, execute_continue, execute_cleanup)

    #no input
    if len(sys.argv) <= 1:
        parser.print_help(sys.stdout)
        return

    namespace = parser.parse_args(sys.argv[1:])
    cstar.output.configure(namespace.verbose)
    namespace.func(namespace)


if __name__ == '__main__':
    main()



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BadSSHHost(Exception):
    pass


class FailedExecution(Exception):
    pass


class BadEnvironmentVariable(Exception):
    pass


class NoHostsSpecified(Exception):
    pass


class HostIsDown(Exception):
    pass


class UnknownHost(Exception):
    def __str__(self):
        return "Unknown host: {}".format(self.args[0])


class BadFileFormatVersion(Exception):
    pass


class FileTooOld(Exception):
    pass


class NoDefaultKeyspace(Exception):
    pass


class ParseException(Exception):
    def __init__(self, line, offset, error):
        msg = line + "\n" + (" " * offset) + "^" + "\n" + error
        super(ParseException, self).__init__(msg)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Argument parsing and related plumbing for the cstarpar command"""

import argparse
import sys
import uuid

import cstar.args
import cstar.command
import cstar.strategy
import cstar.job
import cstar.remote
from cstar.output import msg, error, emph
import cstar.output
import cstar.signalhandler
from cstar.exceptions import NoHostsSpecified


def fallback(*args):
    for arg in args:
        if arg is not None:
            return arg


def parse_job_mode():
    desc = '''cstarpar - A shell tool for executing jobs in parallel. Unlike cstar, cstarpar does not run commands
    remotely, it executes a script locally once for each Cassandra host.'''
    parser = argparse.ArgumentParser(description=desc,
                                     prog="cstarpar")

    cstar.args.add_cstarpar_arguments(parser)
    return parser.parse_args()


def main():
    namespace = parse_job_mode()

    if bool(namespace.seed_host) + bool(namespace.host) + bool(namespace.host_file) != 1:
        error("Exactly one of --seed-host, --host and --host-file must be used", print_traceback=False)

    hosts = None

    if namespace.host_file:
        with open(namespace.host_file) as f:
            hosts = f.readlines()

    if namespace.host:
        hosts = namespace.host

    cstar.output.configure(namespace.verbose)

    with cstar.job.Job() as job:
        env = {}
        job_id = str(uuid.uuid4())
        msg("Job id is", emph(job_id))

        cstar.signalhandler.print_message_and_save_on_sigint(job, job_id)

        job.setup(
            hosts=hosts,
            seeds=namespace.seed_host,
            command=namespace.command,
            job_id=job_id,
            strategy=cstar.strategy.parse(fallback(namespace.strategy, "topology")),
            cluster_parallel=fallback(namespace.cluster_parallel, False),
            dc_parallel=fallback(namespace.dc_parallel, False),
            max_concurrency=namespace.max_concurrency,
            timeout=namespace.timeout,
            env=env,
            stop_after=namespace.stop_after,
            job_runner=cstar.jobrunner.LocalJobRunner,
            key_space=namespace.key_space,
            output_directory=namespace.output_directory,
            ignore_down_nodes=False,
            dc_filter=namespace.dc_filter,
            sleep_on_new_runner=namespace.ssh_pause_time,
            sleep_after_done=namespace.node_done_pause_time,
            ssh_username = namespace.ssh_username,
            ssh_password = namespace.ssh_password,
            ssh_identity_file = namespace.ssh_identity_file)
        job.run()



if __name__ == '__main__':
    main()



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import cstar.strategy
import cstar.progress


class State(object):
    """All the state, all the time.

    This type contains all the different data that is required to find out what node (if any) to run the job on next.

    Somewhat confusingly, the State object needs to keep track of two potentially different topologies of hosts.
    The first topology (original_topology) is the set of hosts that cstar needs to run on. The second topology
    (current_topology) is the current set of nodes in those same clusters. For many cluter operations, those two sets
    will be identical, but cstar needs to be able to support node operations such as spinning down nodes and replacing
    nodes, and in these situations some or all of the nodes towards the end of the execution will be different.

    This type is meant to be used without mutating it."""

    def __init__(
            self, original_topology, strategy, endpoint_mapping, cluster_parallel, dc_parallel,
            max_concurrency=None, progress=None, current_topology=None, stop_after=None,
            ignore_down_nodes=False):
        self.original_topology = original_topology
        self.current_topology = current_topology or original_topology
        self.strategy = strategy
        self.endpoint_mapping = endpoint_mapping
        self.cluster_parallel = cluster_parallel
        self.dc_parallel = dc_parallel
        self.max_concurrency = max_concurrency
        self.progress = progress or cstar.progress.Progress()
        self.stop_after = stop_after
        self.ignore_down_nodes = ignore_down_nodes

    def clone(self):
        return copy.copy(self)

    def with_topology(self, new_topology):
        res = self.clone()
        res.current_topology = new_topology
        return res

    def with_running(self, host):
        return self.with_progress(self.progress.with_running(host))

    def with_done(self, host):
        return self.with_progress(self.progress.with_done(host))

    def with_failed(self, host):
        return self.with_progress(self.progress.with_failed(host))

    def with_progress(self, progress):
        res = self.clone()
        res.progress = progress
        return res

    def find_next_host(self):
        return cstar.strategy.find_next_host(
            topology=self.original_topology, strategy=self.strategy, endpoint_mapping=self.endpoint_mapping,
            max_concurrency=self.max_concurrency, progress=self.progress,
            cluster_parallel=self.cluster_parallel, dc_parallel=self.dc_parallel, stop_after=self.stop_after,
            ignore_down_nodes=self.ignore_down_nodes)

    def is_done(self):
        return (len(self.progress.done) == len(self.original_topology)) or (
            self.stop_after and (len(self.progress.running) + len(self.progress.done) + len(self.progress.failed) >= self.stop_after))

    def is_healthy(self):
        if self.ignore_down_nodes:
            return True
        return len(self.current_topology.without_hosts(self.progress.running).get_down()) == 0

    def get_idle(self):
        return self.current_topology.without_hosts(self.progress.running)

    def __str__(self):
        return """"State(
    original_topology=%s,
    current_topology=%s,
    strategy=%s,
    endpoint_mapping=%s,
    cluster_parallel%s,
    dc_parallel=%s,
    max_concurrency=%s,
    progress=%s)""" % (
            self.original_topology, self.current_topology, self.strategy, self.endpoint_mapping, self.cluster_parallel,
            self.dc_parallel, self.max_concurrency, self.progress)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paramiko.client
import re

from cstar.output import err, debug, msg
from cstar.exceptions import BadSSHHost, BadEnvironmentVariable, NoHostsSpecified
from cstar.executionresult import ExecutionResult

PING_COMMAND = "echo ping"

_alnum_re = re.compile(r"[^a-zA-Z0-9\|_]")


class Remote(object):
    def __init__(self, hostname, ssh_username=None, ssh_password=None, ssh_identity_file=None):
        if hasattr(hostname, "ip"):
            self.hostname = hostname.ip
        else:
            self.hostname = hostname
        if not self.hostname:
            raise NoHostsSpecified("No SSH host specified")
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.ssh_identity_file = ssh_identity_file
        self.client = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def _connect(self):
        if self.client:
            # Ensure underlying client is still a valid open connection
            try:
                stdin, stdout, stderr = self.client.exec_command(PING_COMMAND)
            except (ConnectionResetError, paramiko.ssh_exception.SSHException):
                # ConnectionResetError is raised when a connection was established but then broken
                # paramiko.ssh_exception.SSHException is raised if the connection was known to be broken
                self.client = None

        if not self.client:
            try:
                self.client = paramiko.client.SSHClient()
                pkey = None
                if self.ssh_identity_file != None:
                    pkey = paramiko.RSAKey.from_private_key_file(self.ssh_identity_file, None)
                debug("Username : ", self.ssh_username)
                debug("Id file: ", self.ssh_identity_file)
                self.client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())
                self.client.connect(self.hostname, compress=True, username=self.ssh_username, password=self.ssh_password, pkey=pkey)
            except:
                self.client = None
                raise BadSSHHost("Could not establish an SSH connection to host %s" % (self.hostname,))

    def run_job(self, file, jobid, timeout=None, env={}):
        try:
            self._connect()

            transport = self.client.get_transport()
            session = transport.open_session()
            paramiko.agent.AgentRequestHandler(session)

            dir = ".cstar/remote-jobs/" + jobid

            self.run(("mkdir", "-p", dir))

            self.put_command(file, "%s/job" % (dir,))

            # Manually insert environment into script, since passing env into exec_command leads to it being
            # ignored on most ssh servers. :-(

            for key in env:
                if _alnum_re.search(key):
                    raise BadEnvironmentVariable(key)

            env_str = " ".join(key + "=" + self.escape(value) for key, value in env.items())
            wrapper = r"""#! /bin/sh

    if test -f pid; then
        # We can't wait for things that aren't our children. Loop and sleep. :-(
        while ! test -f status; do
            sleep 10s
        done
        exit
    fi

    %s ./job >stdout 2>stderr &
    echo $! >pid
    wait $!
    echo $? >status
    """ % (env_str,)
            self.write_command(wrapper, "%s/wrapper" % (dir,))

            cmd = """
    cd %s
    nohup ./wrapper
    """ % (self.escape(dir),)

            stdin, stdout, stderr = self.client.exec_command(cmd, timeout=timeout)
            stdout.channel.recv_exit_status()
            real_output = self.read_file(dir + "/stdout")
            real_error = self.read_file(dir + "/stderr")
            real_status = int(self.read_file(dir + "/status"))
            return ExecutionResult(cmd, real_status, real_output, real_error)
        except (ConnectionResetError, paramiko.ssh_exception.SSHException):
            raise BadSSHHost("SSH connection to host %s was reset" % (self.hostname,))

    def get_job_status(self, jobid):
        pass

    def run(self, argv):
        try:
            self._connect()
            cmd = " ".join(self.escape(s) for s in argv)

            stdin, stdout, stderr = self.client.exec_command(cmd)
            status = stdout.channel.recv_exit_status()
            out = stdout.read()
            error = stderr.read()
            if status != 0:
                err("Command %s failed with status %d on host %s" % (cmd, status, self.hostname))
            else:
                debug("Command %s succeeded on host %s, output was %s and %s" %
                      (cmd, self.hostname, str(out, 'utf-8'), str(error, 'utf-8')))
            return ExecutionResult(cmd, status, str(out, 'utf-8'), str(error, 'utf-8'))
        except (ConnectionResetError, paramiko.ssh_exception.SSHException):
            self.client = None
            raise BadSSHHost("SSH connection to host %s was reset" % (self.hostname,))

    @staticmethod
    def escape(input):
        if _alnum_re.search(input):
            return "'" + input.replace("'", r"'\''") + "'"
        return input

    def read_file(self, remotepath):
        self._connect()
        with self.client.open_sftp() as ftp_client:
            with ftp_client.file(remotepath, 'r') as f:
                return str(f.read(), 'utf-8')

    def put_file(self, localpath, remotepath):
        self._connect()
        with self.client.open_sftp() as ftp_client:
            ftp_client.put(localpath, remotepath)

    def put_command(self, localpath, remotepath):
        self._connect()
        with self.client.open_sftp() as ftp_client:
            ftp_client.put(localpath, remotepath)
            ftp_client.chmod(remotepath, 0o755)

    def write_command(self, definition, remotepath):
        self._connect()
        with self.client.open_sftp() as ftp_client:
            with ftp_client.open(remotepath, 'w') as f:
                f.write(definition)
            ftp_client.chmod(remotepath, 0o755)

    def mkdir(self, path):
        self.run("mkdir " + path)

    def close(self):
        if self.client:
            self.client.close()
        self.client = None



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue
import time
import socket
import os
import threading
import functools

import cstar.remote
import cstar.endpoint_mapping
import cstar.topology
import cstar.nodetoolparser
import cstar.state
import cstar.strategy
import cstar.jobrunner
import cstar.jobprinter
import cstar.jobwriter
from cstar.exceptions import BadSSHHost, NoHostsSpecified, HostIsDown, \
    NoDefaultKeyspace, UnknownHost
from cstar.output import msg, debug, emph, info, error

MAX_ATTEMPTS = 3


@functools.lru_cache(None)
def ip_lookup(name):
    return socket.gethostbyname(name)


class Job(object):
    """The class that wires all the business logic together.

    Currently polluted by some business logic of it's own. The job handling and some small snippets of code
    should be moved out of this class.
    """

    def __init__(self):
        self._connections = {}
        self.results = queue.Queue()
        self.handled_finished_jobs = set()
        self.state = None
        self.command = None
        self.job_id = None
        self.timeout = None
        self.env = None
        self.errors = []
        self.do_loop = False
        self.job_runner = None
        self.key_space = None
        self.output_directory = None
        self.is_preheated = False
        self.sleep_on_new_runner = None
        self.sleep_after_done = None
        self.ssh_username = None
        self.ssh_password = None
        self.ssh_identity_file = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

        if exc_type:
            if exc_type == NoHostsSpecified:
                error("No hosts specified")
            elif exc_type in [BadSSHHost, NoDefaultKeyspace, HostIsDown, UnknownHost]:
                error(exc_value)


    def get_cluster_topology(self, seed_nodes):
        count = 0
        tried_hosts = []
        for host in seed_nodes:
            tried_hosts.append(host)
            conn = self._connection(host)

            describe_res = conn.run(("nodetool", "describecluster"))
            topology_res = conn.run(("nodetool", "ring"))

            if (describe_res.status == 0) and (topology_res.status == 0):
                cluster_name = cstar.nodetoolparser.parse_describe_cluster(describe_res.out)
                topology = cstar.nodetoolparser.parse_nodetool_ring(topology_res.out, cluster_name, self.reverse_dns_preheat)
                return topology

            count += 1
            if count >= MAX_ATTEMPTS:
                break
        raise HostIsDown("Could not find any working host while fetching topology. Is Cassandra actually running? Tried the following hosts:",
                         ", ".join(tried_hosts))

    def reverse_dns_preheat(self, ips):
        if self.is_preheated:
            return
        self.is_preheated = True

        def create_lookup_thread(ip):
            return threading.Thread(target=lambda: socket.gethostbyaddr(ip))

        print("Preheating DNS cache")
        threads = [create_lookup_thread(ip) for ip in ips]

        for thread in threads:
            thread.start()

        for thread in threads:
            # Don't wait around forever for slow DNS
            thread.join(1.0)
        print("Preheating done")

    def get_keyspaces(self, conn):
        cfstats_output = conn.run(("nodetool", "cfstats","|","grep","Keyspace"))
        return cstar.nodetoolparser.extract_keyspaces_from_cfstats(cfstats_output.out)

    def get_endpoint_mapping(self, topology):
        count = 0
        tried_hosts = []
        for host in topology.get_up():
            tried_hosts.append(host)
            conn = self._connection(host)

            mappings = []

            if self.key_space:
                keyspaces = [self.key_space]
            else:
                keyspaces = self.get_keyspaces(conn)

            for keyspace in keyspaces:
                if keyspace != "system":
                    debug("Fetching endpoint mapping for keyspace", keyspace)
                    res = conn.run(("nodetool", "describering", keyspace))
                    has_error = False

                    if res.status != 0:
                        has_error = True
                        break
                    describering = cstar.nodetoolparser.parse_nodetool_describering(res.out)
                    range_mapping = cstar.nodetoolparser.convert_describering_to_range_mapping(describering)
                    mappings.append(cstar.endpoint_mapping.parse(range_mapping, topology, lookup=ip_lookup))

            if not has_error:
                return cstar.endpoint_mapping.merge(mappings)

            count += 1
            if count >= MAX_ATTEMPTS:
                break
        raise HostIsDown("Could not find any working host while fetching endpoint mapping. Tried the following hosts:",
                         ", ".join(host.fqdn for host in tried_hosts))

    def setup(self, hosts, seeds, command, job_id, strategy, cluster_parallel, dc_parallel, job_runner,
              max_concurrency, timeout, env, stop_after, key_space, output_directory,
              ignore_down_nodes, dc_filter,
              sleep_on_new_runner, sleep_after_done,
              ssh_username, ssh_password, ssh_identity_file):

        msg("Starting setup")

        msg("Strategy:", cstar.strategy.serialize(strategy))
        msg("DC parallel:", dc_parallel)
        msg("Cluster parallel:", cluster_parallel)

        self.command = command
        self.job_id = job_id
        self.timeout = timeout
        self.env = env
        self.job_runner = job_runner
        self.key_space = key_space
        self.output_directory = output_directory or os.path.expanduser("~/.cstar/jobs/" + job_id)
        self.sleep_on_new_runner = sleep_on_new_runner
        self.sleep_after_done = sleep_after_done
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.ssh_identity_file = ssh_identity_file
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        msg("Loading cluster topology")
        if seeds:
            current_topology = cstar.topology.Topology([])
            for seed in seeds:
                current_topology = current_topology | self.get_cluster_topology((seed,))
            original_topology = current_topology
            if dc_filter:
                original_topology = original_topology.with_dc(dc_filter)
        else:
            current_topology = cstar.topology.Topology()
            hosts_ip_set = set(socket.gethostbyname(host) for host in hosts)
            for raw_host in hosts:
                host = socket.gethostbyname(raw_host)
                if host in current_topology:
                    continue
                current_topology = current_topology | self.get_cluster_topology((host,))
            original_topology = cstar.topology.Topology(host for host in current_topology if host.ip in hosts_ip_set)
        msg("Done loading cluster topology")

        debug("Run on hosts", original_topology)
        debug("in topology", current_topology)

        msg("Generating endpoint mapping")
        if strategy is cstar.strategy.Strategy.TOPOLOGY:
            endpoint_mapping = self.get_endpoint_mapping(current_topology)
            msg("Done generating endpoint mapping")
        else:
            endpoint_mapping = None
            msg("Skipping endpoint mapping because of selected strategy")

        self.state = cstar.state.State(original_topology, strategy, endpoint_mapping, cluster_parallel, dc_parallel,
                                       max_concurrency, current_topology=current_topology, stop_after=stop_after,
                                       ignore_down_nodes=ignore_down_nodes)
        msg("Setup done")

    def update_current_topology(self, skip_nodes=()):
        new_topology = cstar.topology.Topology()
        for cluster in self.state.original_topology.get_clusters():
            seeds = self.state.get_idle().with_cluster(cluster).without_hosts(skip_nodes).get_up()
            # When using the all strategy, all nodes go to running, so we need to pick some node
            seeds = seeds or self.state.current_topology.with_cluster(cluster).get_up()
            new_topology = new_topology | self.get_cluster_topology(seeds)
        self.state = self.state.with_topology(new_topology)

    def wait_for_node_to_return(self, node):
        """Wait until node returns"""
        while True:
            try:
                self.update_current_topology((node,))

                if self.state.is_healthy():
                    break
            except BadSSHHost as e:
                # If the instance used to poll cluster health is down it probably means that machine is rebooting
                # State is then NOT healthy, so continue waiting...
                debug("SSH to %s failed, instance down?" % (node, ), e)
            cstar.jobprinter.print_progress(self.state.original_topology,
                                            self.state.progress,
                                            self.state.current_topology.get_down())
            time.sleep(5)

    def resume(self):
        self.update_current_topology()
        self.resume_on_running_hosts()
        self.run()

    def run(self):
        self.do_loop = True

        cstar.jobwriter.write(self)
        if not self.state.is_healthy():
            raise HostIsDown(
                "Can't run job because hosts are down: " + ", ".join(
                    host.fqdn for host in self.state.current_topology.get_down()))

        while self.do_loop:
            self.schedule_all_runnable_jobs()

            if self.state.is_done():
                self.do_loop = False

            self.wait_for_any_job()

        self.wait_for_all_jobs()
        cstar.jobprinter.print_progress(self.state.original_topology,
                                        self.state.progress,
                                        self.state.current_topology.get_down())
        self.print_outcome()

    def resume_on_running_hosts(self):
        for host in self.state.progress.running:
            debug("Resume on host", host.fqdn)
            threading.Thread(target=self.job_runner(self, host, self.ssh_username, self.ssh_password, self.ssh_identity_file),
                             name="cstar %s" % host.fqdn).start()
            time.sleep(self.sleep_on_new_runner)

    def print_outcome(self):
        if self.state.is_done() and not self.errors:
            if len(self.state.progress.done) == self.state.stop_after:
                cstar.jobwriter.write(self)
                msg("Job", self.job_id, "successfully ran on", self.state.stop_after, "hosts.\nTo finish the job, run",
                    emph("cstar continue %s" % (self.job_id,)))

            msg("Job", self.job_id, "finished successfully")
        else:
            msg("Job", self.job_id, "finished with errors.\n"
                                    "%s nodes finished successfully\n"
                                    "%s nodes had errors\n"
                                    "%s nodes didn't start executing"
                                    % (len(self.state.progress.done),
                                       len(self.state.progress.failed),
                                       len(self.state.original_topology) - len(self.state.progress.done) - len(self.state.progress.failed)))

    def wait_for_all_jobs(self):
        while self.state.progress.running:
            host, result = self.results.get()
            self.handle_finished_job(host, result)

    def wait_for_any_job(self):
        if self.do_loop:
            host, result = self.results.get(timeout=self.timeout)
            self.handle_finished_job(host, result)
            self.wait_for_node_to_return(host)

    def handle_finished_job(self, host, result):
        if result.status != 0:
            self.errors.append((host, result))
            self.state = self.state.with_failed(host)
            msg("Failure on host", host.fqdn)
            if result.out:
                msg("stdout:", result.out)
            if result.err:
                msg("stderr:", result.err)
            self.do_loop = False
        else:
            self.state = self.state.with_done(host)
            info("Host %s finished successfully" % (host.fqdn,))
            if result.out:
                info("stdout:", result.out, sep="\n")
            if result.err:
                info("stderr:", result.err)
            if self.sleep_after_done:
                debug("Sleeping %d seconds..." % self.sleep_after_done)
                time.sleep(self.sleep_after_done)
        cstar.jobwriter.write(self)
        # Signal the jobrunner that it can delete the remote job files and terminate.
        self.handled_finished_jobs.add(host)

    def schedule_all_runnable_jobs(self):
        while True:
            next_host = self.state.find_next_host()
            if not next_host:
                if not self.state.progress.running:
                    self.do_loop = False
                break
            if (not next_host.is_up) and self.state.ignore_down_nodes:
                self.state = self.state.with_done(next_host)
            else:
                self.state = self.state.with_running(next_host)
                self.schedule_job(next_host)
            cstar.jobwriter.write(self)
            cstar.jobprinter.print_progress(self.state.original_topology,
                                            self.state.progress,
                                            self.state.current_topology.get_down())

    def schedule_job(self, host):
        debug("Running on host", host.fqdn)
        threading.Thread(target=self.job_runner(self, host, self.ssh_username, self.ssh_password, self.ssh_identity_file),
                         name="cstar %s" % host.fqdn).start()
        time.sleep(self.sleep_on_new_runner)

    def _connection(self, host):
        if host not in self._connections:
            self._connections[host] = cstar.remote.Remote(host, self.ssh_username, self.ssh_password, self.ssh_identity_file)
        return self._connections[host]

    def close(self):
        for name, conn in self._connections.items():
            if conn:
                conn.close()
        self._connections = {}



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import curses
import traceback

_print_debug = False
_print_info = False
_last_message_was_topology = False
_last_topology_lines = None
_term_setup = False

CURSOR_UP = 'CURSOR_UP'
CLEAR_LINE = 'CLEAR_LINE'
SET_BOLD = 'SET_BOLD'
SET_NORMAL = 'SET_NORMAL'
CODES = {CURSOR_UP: 'cuu1', CLEAR_LINE: 'el', SET_BOLD: 'bold', SET_NORMAL: 'sgr0'}


def get_termstr(action):
    global _term_setup
    if not _term_setup:
        curses.setupterm()
        _term_setup = True
    return curses.tigetstr(CODES[action]).decode('utf-8')


def configure(verbosity):
    global _print_debug, _print_info
    _print_debug = verbosity >= 2
    _print_info = verbosity >= 1


def msg(*args, **kwargs):
    global _last_message_was_topology
    _last_message_was_topology = False
    print(*args, **kwargs)


def print_topology(top):
    global _last_message_was_topology, _last_topology_lines
    if _last_message_was_topology:
        top = (get_termstr(CLEAR_LINE) + get_termstr(CURSOR_UP)) * _last_topology_lines + get_termstr(CLEAR_LINE) + top

    print(top)
    _last_topology_lines = top.count('\n') + 1
    _last_message_was_topology = True


def info(*args, **kwargs):
    global _last_message_was_topology
    if _print_info:
        _last_message_was_topology = False
        print(*args, **kwargs)


def err(*args, **kwargs):
    global _last_message_was_topology
    _last_message_was_topology = False
    print(*args, file=sys.stderr, **kwargs)


def debug(*args, **kwargs):
    global _last_message_was_topology
    if _print_debug:
        _last_message_was_topology = False
        print(*args, file=sys.stderr, **kwargs)


def warn(*args, **kwargs):
    global _last_message_was_topology
    _last_message_was_topology = False
    print(*args, file=sys.stderr, **kwargs)


def emph(msg):
    return get_termstr(SET_BOLD) + msg + get_termstr(SET_NORMAL)


def error(*args, code=1, print_traceback=True):
    if print_traceback:
        traceback.print_exc()
    err(emph("Error:"), *args)
    sys.exit(code)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parse the output of the converted nodetool describering output"""

import socket

import collections


def parse(range_mapping, topology, lookup=socket.gethostbyname):
    return _topology_map(_dns_map(_parse_range_mapping(range_mapping), lookup), topology)


def merge(mappings):
    res = collections.defaultdict(lambda: set())
    for mapping in mappings:
        for host, friends in mapping.items():
            res[host] = res[host] | friends
    return res


def _parse_range_mapping(ranges):
    mapping = {}
    for range in ranges:
        for host1 in range.get("endpoints"):
            for host2 in range.get("endpoints"):
                if host1 != host2:
                    mapping.setdefault(host1, set()).add(host2)
    return mapping


def _dns_map(mapping, lookup=socket.gethostbyname):
    return dict(
        (lookup(raw_host), set(lookup(raw_friend) for raw_friend in raw_friends))
        for raw_host, raw_friends in mapping.items())


def _topology_map(mapping, topology):
    res = {}
    # Fallback for case with no overlap
    for host in topology:
        res[host] = set()
    for raw_host, raw_friends in mapping.items():
        friends = (topology.get_host(raw_friend) for raw_friend in raw_friends)
        host = topology.get_host(raw_host)
        filtered_friends = set(friend for friend in friends if friend.dc == host.dc)
        res[host] = filtered_friends
    return res



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Locate and parse cstar commands"""

import json
import os
import re
import functools
from collections import namedtuple
from stat import S_ISREG
from pkg_resources import resource_filename

from cstar.exceptions import BadEnvironmentVariable
from cstar.output import warn

_property_re = re.compile(r"^# C\*\s*([^\s:]+)\s*:\s*(.*)\s*$", re.MULTILINE)
_env_re = re.compile("[^a-zA-Z0-9_]")

Command = namedtuple("Command", "name file strategy cluster_parallel dc_parallel arguments description")


class Argument(object):
    """Represents an argument for a command"""

    def __init__(self, name, option, description, required=False, default=None):
        for name, val in locals().items():
            if name != "self":
                self.__setattr__(name, val)
        if _env_re.search(self.name):
            raise BadEnvironmentVariable(self.name)


def load(command_name):
    """Loads a command"""
    file = _search(command_name)
    with open(file) as f:
        return _parse(command_name, file, f.read())


def _parse(command_name, filename, definition):
    """Parses a command"""
    cluster_parallel = None
    dc_parallel = None
    description = None
    strategy = "topology"
    arguments = []

    truth_lookup = {"true": True, "false": False}

    for line in definition.split('\n'):

        if not line or line[0] != "#":
            break

        match = _property_re.match(line)
        if match:
            name = match.group(1)
            value = match.group(2)
            if name == "cluster-parallel":
                cluster_parallel = truth_lookup[value.lower()]
            elif name == "dc-parallel":
                dc_parallel = truth_lookup[value.lower()]
            elif name == "description":
                description = value
            elif name == "strategy":
                strategy = value
            elif name == "argument":
                arguments.append(Argument(**json.loads(value)))
            else:
                warn("Ignoring unknown property %s while parsing %s" % (name, filename))

    return Command(name=command_name, file=filename,
                   cluster_parallel=cluster_parallel,
                   dc_parallel=dc_parallel,
                   description=description,
                   strategy=strategy,
                   arguments=arguments)


def _search_path():
    return os.path.expanduser('~/.cstar/commands'), '/etc/cstar/commands', resource_filename('cstar.resources', 'commands')


def _stat_is_reg(stat_output):
    return S_ISREG(stat_output.st_mode)


def _search(name, listdir=os.listdir, stat=os.stat, check_is_file=_stat_is_reg):
    """Returns a the filename for a given command"""
    if "/" not in name:
        listing = _list(listdir, stat, check_is_file)
        if name in listing:
            return listing[name]
    if check_is_file(stat(name)):
        return name
    raise FileNotFoundError("Failed to find definition for command %s" % (name,))


def list(listdir=os.listdir, stat=os.stat, check_is_file=_stat_is_reg):
    return _list(listdir, stat, check_is_file).keys()


@functools.lru_cache(None)
def _list(listdir, stat, check_is_file):
    """Returns a list of the names of all available commands"""
    res = {}
    for dir in _search_path():
        try:
            for filename in listdir(dir):
                full_name = os.path.join(dir, filename)
                try:
                    if not check_is_file(stat(full_name)):
                        continue
                except FileNotFoundError:
                    continue
                if filename.endswith("~") or filename.startswith("#"):
                    continue

                if "." in filename:
                    prefix = re.sub(r"\..*", "", filename)
                else:
                    prefix = filename
                if prefix not in res:
                    res[prefix] = full_name
        except FileNotFoundError:
            pass
    return res



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cstar.remote
import cstar.jobwriter
import subprocess
import shlex
import os
import time

from cstar.executionresult import ExecutionResult


def save_output(job, host, result):
    output_directory = job.output_directory + "/" + host.fqdn
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    with open(output_directory + "/out", 'w') as f:
        f.write(result.out)
    with open(output_directory + "/err", 'w') as f:
        f.write(result.err)
    with open(output_directory + "/status", 'w') as f:
        f.write(str(result.status))


class RemoteJobRunner(object):
    """Job runners are responsible for running a command to completion for a specific host and reporting back the
     status.

    This is where network outage handling logic will eventually go.

    When a RemoteJobRunner is rerun twice for the same host in the same job, the command will still only be executed
    once.

    This JobRunner is used by cstar
    """

    def __init__(self, job, host, ssh_username=None, ssh_password=None, ssh_identity_file=None):
        self.job = job
        self.host = host
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.ssh_identity_file = ssh_identity_file

    def __call__(self):
        with cstar.remote.Remote(self.host, self.ssh_username, self.ssh_password, self.ssh_identity_file) as conn:
            result = conn.run_job(self.job.command, self.job.job_id, self.job.timeout, self.job.env)
            self.job.results.put((self.host, result))  # This signals the main thread that the job completed.
            save_output(self.job, self.host, result)
            # We have to make sure that the local job file is updated before the remote
            # job data is deleted. We need to syncronize with the main thread.
            while self.host not in self.job.handled_finished_jobs:
                time.sleep(1)
            conn.run(("rm", "-rf", ".cstar/remote-jobs/" + self.job.job_id))
        return result


class LocalJobRunner(object):
    """Job runners are responsible for running a command to completion for a specific host and reporting back the
     status. This one executes the command locally.

    When a LocalJobRunner is rerun twice for the same host in the same job, the command will be executed twice.

     This JobRunner is used by cstarpar
    """

    def __init__(self, job, host, ssh_username=None, ssh_password=None, ssh_identity_file=None):
        self.job = job
        self.host = host

    def __call__(self):
        replaced = self.job.command.replace("{}", self.host.ip)

        # Use subprocess.Popen because subprocess.run is not available in Python 3.4 used by Ubuntu Trusty
        proc = subprocess.Popen(shlex.split(replaced), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            outs, errs = proc.communicate()
        except subprocess.TimeoutExpired:
            # The child process is not killed in case of timeout
            proc.kill()
            outs, errs = proc.communicate()
        result = ExecutionResult(replaced, proc.returncode, str(outs, 'utf-8'), str(errs, 'utf-8'))

        save_output(self.job, self.host, result)
        self.job.results.put((self.host, result))
        return result



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import cstar.exceptions


class ExecutionResult(namedtuple("ExecutionResult", "command status out err")):

    def out_or_raise(self):
        if self.status != 0:
            raise cstar.exceptions.FailedExecution('Command "%s" failed with status %d and error message %s' %
                                                   (self.command, self.status, self.err))
        return self.out



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from cstar.exceptions import HostIsDown


class Strategy(Enum):
    """The three concurrency strategies available in cstar"""
    ONE = 1
    TOPOLOGY = 2
    ALL = 3


def parse(text):
    """Convert a strategy string to a strategy object"""
    return {
        "one": Strategy.ONE,
        "topology": Strategy.TOPOLOGY,
        "all": Strategy.ALL
    }[text]


def serialize(s):
    """Convert a strategy object to a strategy string"""
    return {
        Strategy.ONE: "one",
        Strategy.TOPOLOGY: "topology",
        Strategy.ALL: "all"
    }[s]


def find_next_host(strategy, topology, endpoint_mapping, progress, cluster_parallel, dc_parallel, max_concurrency,
                   stop_after, ignore_down_nodes):
    """Entry point for figuring out which host to do things on next."""

    if stop_after and ((len(progress.running) + len(progress.done) + len(progress.failed)) >= stop_after):
        return None

    remaining = topology.without_hosts(progress.done).without_hosts(progress.running).without_hosts(progress.failed)
    if progress.running and not cluster_parallel:
        remaining = remaining.with_cluster(next(iter(progress.running)).cluster)

    if progress.running and not dc_parallel:
        remaining = remaining.with_dc(next(iter(progress.running)).dc)

    if not remaining:
        return None

    if max_concurrency and (len(progress.running) >= max_concurrency):
        return None

    if not ignore_down_nodes:
        for host in remaining:
            if not host.is_up:
                raise HostIsDown(host)

    return _strategy_mapping[strategy](remaining, endpoint_mapping, progress.running)


def _all_find_next_host(remaining, endpoint_mapping, running):
    return remaining.first()


def _one_find_next_host(remaining, endpoint_mapping, running):
    if running:
        return None
    return remaining.first()


def _topology_find_next_host(remaining, endpoint_mapping, running):
    for h in running:
        for next in endpoint_mapping[h]:
            remaining = remaining.without_host(next)
    return remaining.first()


_strategy_mapping = {
    Strategy.ONE: _one_find_next_host,
    Strategy.TOPOLOGY: _topology_find_next_host,
    Strategy.ALL: _all_find_next_host}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
from setuptools.command.install import install
import os
import sys

# circleci.py version
VERSION = '0.5.1'

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)

setup(name='cstar',
      version=VERSION,
      author='Spotify',
      author_email='rebase-squad@spotify.com',
      url='https://github.com/spotify/cstar',
      description='Apache Cassandra cluster orchestration tool for the command line',
      license='Apache-2.0',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.6',
          'Topic :: Database'
      ],
      install_requires=('paramiko'),
      python_requires='>=3',
      packages=('cstar', 'cstar.nodetoolparser', 'cstar.resources'),
      package_data={'cstar.resources': ['commands/*']},
      test_suite='tests',
      entry_points={
          'console_scripts': [
              'cstar=cstar.cstarcli:main',
              'cstarpar=cstar.cstarparcli:main',
          ]},
      cmdclass={
          'verify': VerifyVersionCommand,
      }
)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


class Progress(object):
    """The progress of a job.
    Progress is expressed in terms of what state each node is in.
    This type is meant to be used without mutating it."""

    def __init__(self, done=None, running=None, failed=None):
        self.done = set(done or [])
        self.running = set(running or [])
        self.failed = set(failed or [])

    def clone(self):
        return copy.copy(self)

    def with_running(self, host):
        res = self.clone()
        res.running.add(host)
        return res

    def with_done(self, host):
        res = self.clone()
        res.done.add(host)
        if host in res.running:
            res.running.remove(host)
        return res

    def with_failed(self, host):
        res = self.clone()
        res.failed.add(host)
        if host in res.running:
            res.running.remove(host)
        return res

    def __str__(self):
        return """"Progress(
    done=%s,
    running=%s,
    failed=%s)""" % (self.done, self.running, self.failed)



# Copyright 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cstar.job
import cstar.jobreader
import cstar.exceptions
import os
import shutil

from cstar.output import msg


def cleanup(max_days, listdir=os.listdir, jobread=cstar.jobreader.read, delete=shutil.rmtree):
    job_dir = os.path.expanduser('~/.cstar/jobs')
    for job_id in listdir(job_dir):
        try:
            jobread(cstar.job.Job(), job_id, stop_after=None, max_days=max_days, endpoint_mapper=lambda x: None)
        except Exception:
            msg("Removing job", job_id)
            full_name = os.path.join(job_dir, job_id)
            delete(full_name)



# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import os

import cstar.jobrunner
import cstar.jobwriter
import cstar.strategy
import cstar.topology
from cstar.exceptions import BadFileFormatVersion, FileTooOld
from cstar.output import debug


def read(job, job_id, stop_after, output_directory=None, max_days=7, endpoint_mapper=None, retry=False):
    output_directory = output_directory or os.path.expanduser("~/.cstar/jobs/" + job_id)
    file = output_directory + "/job.json"

    if not endpoint_mapper:
        endpoint_mapper = job.get_endpoint_mapping

    with open(file) as f:
        return _parse(f.read(), file, output_directory, job, job_id, stop_after, max_days, endpoint_mapper, retry)
    return job


def _parse(input, file, output_directory, job, job_id, stop_after, max_days, endpoint_mapper, retry=False):
    data = json.loads(input)

    if 'version' not in data:
        raise BadFileFormatVersion("Incompatible file format version, wanted %d" %
                                   (cstar.jobwriter.FILE_FORMAT_VERSION,))
    if data['version'] != cstar.jobwriter.FILE_FORMAT_VERSION:
        raise BadFileFormatVersion("Incompatible file format version, wanted %d but %s is of version %d" %
                                   (cstar.jobwriter.FILE_FORMAT_VERSION, file, data['version']))

    creation_time = datetime.datetime.utcfromtimestamp(data["creation_timestamp"])
    age = (datetime.datetime.utcnow() - creation_time).days
    if age > max_days:
        raise FileTooOld(("Job created %d days ago, which is more than the current maximum age of %d. " +
                          "Use --max-job-age %d if you really want to run this job.") % (age, max_days, age + 1))

    state = data['state']
    job.command = data['command']
    job.job_id = job_id
    job.timeout = data['timeout']
    job.env = data['env']
    job.job_runner = getattr(cstar.jobrunner, data["job_runner"])
    job.key_space = data['key_space'] if 'key_space' in data else None
    job.output_directory = output_directory
    job.sleep_on_new_runner = data['sleep_on_new_runner']

    strategy = cstar.strategy.parse(state['strategy'])
    cluster_parallel = state['cluster_parallel']
    dc_parallel = state['dc_parallel']
    max_concurrency = state['max_concurrency']

    progress = cstar.progress.Progress(
        running=[cstar.topology.Host(*arr) for arr in state['progress']['running']],
        done=[cstar.topology.Host(*arr) for arr in state['progress']['done']],
        failed=[cstar.topology.Host(*arr) for arr in state['progress']['failed']])

    if retry==True:
        progress.failed = set([])
    
    original_topology = cstar.topology.Topology(cstar.topology.Host(*arr) for arr in state['original_topology'])
    current_topology = cstar.topology.Topology(cstar.topology.Host(*arr) for arr in state['current_topology'])

    debug("Run on hosts", original_topology)
    debug("in topology", current_topology)

    if strategy is cstar.strategy.Strategy.TOPOLOGY:
        endpoint_mapping = endpoint_mapper(original_topology)
    else:
        endpoint_mapping = None

    job.state = cstar.state.State(
        original_topology=original_topology,
        strategy=strategy,
        endpoint_mapping=endpoint_mapping,
        cluster_parallel=cluster_parallel,
        dc_parallel=dc_parallel,
        max_concurrency=max_concurrency,
        current_topology=current_topology,
        stop_after=stop_after,
        progress=progress,
        ignore_down_nodes=state['ignore_down_nodes'])
